--
-- attachableCylinderedParts
-- Specialization for cylindered parts in various vehicles
--
-- @author John Deere 6930
-- @date 05/05/2020
--

--[[XML
<cylindered>
	<attachableCylinderedParts>
        <part [referenceFrame="" referencePoint=""] attacherJointIndices=""      referenceIdentName="" />
        <part node=""          inputAttacherJointIndices="" referenceIdentName="" />
    </attachableCylinderedParts>
</cylindered>

]]

AttachableCylinderedParts = {};
AttachableCylinderedParts.modDir = g_currentModDirectory;
AttachableCylinderedParts.MOD_NAME = g_currentModName;

function AttachableCylinderedParts.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Cylindered, specializations);
end;

function AttachableCylinderedParts.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "loadAttachablePart",      AttachableCylinderedParts.loadAttachablePart);
	SpecializationUtil.registerFunction(vehicleType, "getArePartsValid",           AttachableCylinderedParts.getArePartsValid);
	SpecializationUtil.registerFunction(vehicleType, "connectMovingParts",      AttachableCylinderedParts.connectMovingParts);
	SpecializationUtil.registerFunction(vehicleType, "disconnectMovingParts",   AttachableCylinderedParts.disconnectMovingParts);
end;

function AttachableCylinderedParts.registerOverwrittenFunctions(vehicleType)
end;

function AttachableCylinderedParts.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", AttachableCylinderedParts);
    SpecializationUtil.registerEventListener(vehicleType, "onPostAttachImplement", AttachableCylinderedParts);
    SpecializationUtil.registerEventListener(vehicleType, "onPreDetachImplement", AttachableCylinderedParts);
end;

function AttachableCylinderedParts:onLoad(savegame)
    self.spec_attachableCylinderedParts = self[string.format("spec_%s.attachableCylinderedParts", AttachableCylinderedParts.MOD_NAME)];
    local spec = self.spec_attachableCylinderedParts;
	
	spec.attachableCylinderedParts = {}
    local i=0
    while true do
        local baseName = string.format("vehicle.cylindered.attachableCylinderedParts.part(%d)", i);
        if not hasXMLProperty(self.xmlFile, baseName) then
            break;
        end;
        
        local entry = {};
        if self:loadAttachablePart(self.xmlFile, baseName, entry) then
            table.insert(spec.attachableCylinderedParts, entry);
        end;
		i = i + 1;
	end;
end;

function AttachableCylinderedParts:loadAttachablePart(xmlFile, key, entry)
    local node = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key.."#node"), self.i3dMappings);
    if node ~= nil then
        entry.node = node;
    else
        local nodeValid = false;
        local referenceFrame = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key.."#referenceFrame"), self.i3dMappings);
        if referenceFrame ~= nil then
            entry.referenceFrame = referenceFrame;
            nodeValid = true;
        end;
        entry.referencePoint = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key.."#referencePoint"), self.i3dMappings)
        if referencePoint ~= nil then
            entry.referencePoint = referencePoint;
            nodeValid = true;
        end;
        if not nodeValid then
            g_logManager:xmlWarning(self.configFileName, "AttachableCylinderedPart '%s' no valid reference given!", key);
            return false;
        end;
    end;
    entry.connected = false;
    entry.connectedPart = nil;
    
    local referenceIdentName = getXMLString(xmlFile, key.."#referenceIdentName");
    if referenceIdentName ~= nil and referenceIdentName ~= "" then
        entry.referenceIdentName = referenceIdentName;            
    else
        g_logManager:xmlWarning(self.configFileName, "AttachableCylinderedPart '%s' no referenceIdentName given!", key);
        return false;
    end;
    
    local jointsValid = false;
    --load AttacherJoints
    entry.attacherJointIndices = {};
    local indices = StringUtil.getVectorNFromString(getXMLString(xmlFile, key.. "#attacherJointIndices"));
    if indices ~= nil then
        for _,  index in ipairs(indices) do
            if self:getAttacherJointByJointDescIndex(index) ~= nil then
                entry.attacherJointIndices[index] = true;
                jointsValid = true;
            end;
        end;          
    end;
    
    --load inputAttacherJoints
    entry.inputAttacherJointIndices = {};
    local indices = StringUtil.getVectorNFromString(getXMLString(xmlFile, key.. "#inputAttacherJointIndices"));
    if indices ~= nil then
        for _,  index in ipairs(indices) do
            if self:getInputAttacherJointByJointDescIndex(index) ~= nil then
                entry.inputAttacherJointIndices[index] = true;
                jointsValid = true;
            end;
        end;          
    end;
    
    if jointsValid then
        return true;
    else
        g_logManager:xmlWarning(self.configFileName, "AttachableCylinderedPart '%s' at least one inputAttacherJoint or attacherJoint must be given!", key);
        return false;
    end;
end;

function AttachableCylinderedParts:onPostAttachImplement(attachableObject, inputJointDescIndex, jointDescIndex)
    local spec = self.spec_attachableCylinderedParts;    
    if self:getArePartsValid(attachableObject) then
        local attachedParts = attachableObject.spec_attachableCylinderedParts; 
        for k,part in pairs(spec.attachableCylinderedParts) do
            if not part.connected then
                if part.attacherJointIndices[jointDescIndex] then
                    for k,partAttached in pairs(attachedParts.attachableCylinderedParts) do
                        if not part.connected then
                            if partAttached.inputAttacherJointIndices[inputJointDescIndex] then
                                if part.referenceIdentName == partAttached.referenceIdentName then
                                    part.connected = self:connectMovingParts(part, partAttached, attachableObject);
                                end;
                            end;
                        end;
                    end;
                end;
            end;
        end;
    end;
end;

function AttachableCylinderedParts:onPreDetachImplement(implement)    
    local spec = self.spec_attachableCylinderedParts;    
    if self:getArePartsValid(implement.object) then
        for k,part in pairs(spec.attachableCylinderedParts) do
            if part.connected then
                part.connected = not self:disconnectMovingParts(part, implement.object);
            end;
        end;
    end;
end;

function AttachableCylinderedParts:getArePartsValid(attachableObject)
    return #self.spec_attachableCylinderedParts.attachableCylinderedParts > 0 and attachableObject.spec_attachableCylinderedParts ~= nil and #attachableObject.spec_attachableCylinderedParts.attachableCylinderedParts > 0;
end;

function AttachableCylinderedParts:connectMovingParts(part, partAttached, attachedVehicle)
    local retVal = false;
    part.connectedPart = nil;
    if part.node ~= nil then
        --get references
        if partAttached.referenceFrame ~= nil then
            --replace referenceFrame
            local movingPart = self:getMovingPartByNode(part.node);
            if movingPart ~= nil then
                movingPart.referenceFrameBackUp = movingPart.referenceFrame;
                movingPart.referenceFrame = partAttached.referenceFrame;
                retVal = true;
                part.connectedPart = partAttached;
            end;
        elseif partAttached.referencePoint ~= nil then
            --replace referencePoint
            local movingPart = self:getMovingPartByNode(part.node);
            if movingPart ~= nil then
                movingPart.referencePointBackUp = movingPart.referencePoint;
                movingPart.referencePoint = partAttached.referencePoint;
                retVal = true;
                part.connectedPart = partAttached;
            end;
        end;
    elseif part.referenceFrame ~= nil then
        --replace referenceFrame
        if partAttached.node ~= nil then
            local movingPartAttached = attachedVehicle:getMovingPartByNode(partAttached.node);
            if movingPartAttached ~= nil then
                movingPartAttached.referenceFrameBackUp = movingPartAttached.referenceFrame;
                movingPartAttached.referenceFrame = part.referenceFrame;
                retVal = true;
                part.connectedPart = partAttached;
            end;
        end;
    elseif part.referencePoint ~= nil then
        --replace referencePoint        
        if partAttached.node ~= nil then
            local movingPartAttached = attachedVehicle:getMovingPartByNode(partAttached.node);
            if movingPartAttached ~= nil then
                movingPartAttached.referencePointBackUp = movingPartAttached.referencePoint;
                movingPartAttached.referencePoint = part.referencePoint;
                retVal = true;
                part.connectedPart = partAttached;
            end;
        end;
    end;
    return retVal;
end;

function AttachableCylinderedParts:disconnectMovingParts(part, attachedVehicle)
    local retVal = false;
    if part.node ~= nil then
        --get references
        if part.connectedPart.referenceFrame ~= nil then
            --replace referenceFrame
            local movingPart = self:getMovingPartByNode(part.node);
            if movingPart ~= nil then
                movingPart.referenceFrame = movingPart.referenceFrameBackUp;
                retVal = true;
            end;
        elseif part.connectedPart.referencePoint ~= nil then
            --replace referencePoint
            local movingPart = self:getMovingPartByNode(part.node);
            if movingPart ~= nil then
                movingPart.referencePoint = movingPart.referencePointBackUp;
                retVal = true;
            end;
        end;
    elseif part.referenceFrame ~= nil then
        --replace referenceFrame
        if part.connectedPart.node ~= nil then
            local movingPartAttached = attachedVehicle:getMovingPartByNode(part.connectedPart.node);
            if movingPartAttached ~= nil then
                movingPartAttached.referenceFrame = movingPartAttached.referenceFrameBackUp;
                retVal = true;
            end;
        end;
    elseif part.referencePoint ~= nil then
        --replace referencePoint        
        if part.connectedPart.node ~= nil then
            local movingPartAttached = attachedVehicle:getMovingPartByNode(part.connectedPart.node);
            if movingPartAttached ~= nil then
                movingPartAttached.referencePoint = movingPartAttached.referencePointBackUp;
                retVal = true;
            end;
        end;
    end;
    return retVal;
end;

